"""
  This is one of 2 files required to run the neural chatbot in nest.  The other 
  is SimpleChatbotAgentNest.py. This file contains the interface (world).  
  SimpleChatbotAgentNest.py   contains the chatbot (agent).  You need to run 
  SimpleChatbotAgentNest.py in one shell before you run 
  ChatBotInterfaceNext.py in another shell.

  When you run this file (i.e. ChatBotInterfaceNest.py) an 
      "enter a question" 
  prompt should appear for you to enter a question.  Currently, you can enter
  one of 3 queries:
      Where is the yellow banana?
      Where is the red apple?
      What shape is the diamond?

  The language, i.e. the sentences the agent can use to answer the queries, 
  are defined in chatbotLanguage.txt. The answers to the above queries are 
  programmed/defined in LanguageGenerator.py.

  A query entered at the interface activates a query cell in the agent 
  (chatbot), the agent processes the query and sends an answer back to the 
  interface via a set of spikes.

Communication between interface and agent is acheived through text files:  query.txt contains information from the environment (interface) for
the agent; outputWordsSpikes.txt and AgentQuit.txt contains information from the agent for the environment.

query.txt contains the query (i.e. in the form of an integer) from the user

outputWordsSpikes.txt contains the spikes from the agent, i.e. the spike times and the word neurons that fired.  Each word neuron
represents a different word (the words represented are shown in OutputWords.txt). Therefore, the interface builds the response (sentence) using
the words in OutputWords.txt and the spikes, and spike times, from the word neurons.

the interface will quit if AgentQuit.txt appears, i.e. when the agent (chatbot) quits

"""

import time
import numpy as np
import os.path
from generatorBaseClass import GeneratorBaseClass


Words =[]       #stores the words from file "OutputWords.txt".  These are all of the different words comprising the
                #language (sentences) in chatbotLanguage.txt
currentLine = 0 ##records where in the file outputWordsSpikes.txt, i.e. the line, new spikes will appear.
wordSpikeTimes=[]  ##stores the spike times of the word neurons that fired
wordNeuronsFired=[] ##stores the word neurons that fired
ReceivedFullResponse=0 


def sendQuery(query):
    ##this sends the query as an integer via query.txt
    nestOutputFileHandle = open("query.txt",'w') 
    nestOutputFileHandle.write(str(query))  
    nestOutputFileHandle.close()
    print ("sent Query, waiting for response....")
    

def GetQueryFromUser():
    #this asks the user for a question and sends it to the agent (chatbot)

    validquestion=0
    
    while(validquestion==0):
        query=input("Enter a question: ")
        if(HasAgentQuit()!=1): # if a 1, agent has quit so do not send
            if(query=="where is the yellow banana?"):
                sendQuery(0) # a 0 corresponds to "where is the yellow banana?"
                validquestion=1
            elif(query=="where is the red apple?"):
                sendQuery(1) # a 1 corresponds to "where is the red apple?"
                validquestion=1
            elif(query=="what shape is the diamond?"):
                sendQuery(2) # a 2 corresponds to "what shape is the diamond?"
                validquestion=1            
            else:
                print ("not a valid Question, try again")
        else:
            break

def getAgentResponse(generator):
    ##this retrieves the agent's response to the current query.  In other words, it extracts from file outputWordsSpikes.txt the new spikes
    ##fired by the word population (i.e. within the last time step (currentSimulationTime-30<currentSimulationTime))
    
    
    global wordSpikeTimes
    global wordNeuronsFired
    global Words
    
    global currentLine ##records where in the file outputWordsSpikes.txt, i.e. the line, new spikes will appear.
    global ReceivedFullResponse

    if (os.path.isfile("outputWordsSpikes.txt")):
        ##obtain the spikes from the agent (chatbot)
        fileHandle = open("outputWordsSpikes.txt",'r')
        ##check for new spikes in the file, i.e. within the last time step
        lineCount = 0
        for inpLine in fileHandle:
            lineCount += 1
            if (lineCount > currentLine): 
                splitInp = inpLine.split()
                if(len(splitInp)>1):
                    wordSpikeTimes+=[float(splitInp[0])]
                    wordNeuronsFired+=[int(splitInp[1])]
                    if(Words[int(splitInp[1])]=="."):
                        #if the interface has received the full response from
                        #the agent, i.e. the word neuron associated to a full
                        #stop has fired, print out the sentence received and
                        #clear the stored spike times and neurons that fired
                        #ready for the next answer.
                        generator.printSentenceLive2(wordNeuronsFired,wordSpikeTimes,Words)
                        wordSpikeTimes=[]
                        wordNeuronsFired=[]
                        ReceivedFullResponse=1  #agent has sent full response.
                
                
        currentLine=lineCount
        fileHandle.close()

def HasAgentQuit():
    ##checks for file AgentQuit.txt
    ##if file exists, the agent has quit
    if (os.path.isfile("AgentQuit.txt")):
        fileHandle = open("AgentQuit.txt",'r')
        Quit=1

    else:
        Quit=0  #there is no file yet
    return Quit


###MAIN

generator = GeneratorBaseClass("","","","","","","OutputWords.txt")  #we do not want to setup a neural language generator here, we
                                                    #only want to use some of the class functions so we can leave some of the parameters blank.
                                                    #However, we need to pass the file that contains all of the different words comprising  
                                                    #the language (sentences) in file chatbotLanguage.txt
##Read in all of the different words used by the agent
Words=generator.RetrieveWords()



#initialise the system and ask user for the first question
GetQueryFromUser()

#user enters a query, program listens for/retrieves the agent's responses overtime. When the interface receives a full response, ask for next query.
while True: 
    
    if(ReceivedFullResponse==1): #if we have received a full response from the agent
        GetQueryFromUser()      #get the next query from user
        ReceivedFullResponse=0
    else:
        getAgentResponse(generator)  #listen for agent responses (i.e. new word spikes fired within the last time step
                                    #(currentSimulationTime-30<currentSimulationTime)

    if(HasAgentQuit()==1):
        print ("program quit as agent (chatbot) has quit")
        break
    
    time.sleep(0.5) # wait for an arbitatry amount of time i.e. wait for the agent to process a response.
