""" 
This builds a language generator topology using the generatorBaseClass.
This subclass should be used with Language file chatbotLanguage.txt.
This system can generate the following sentences:  

The yellow banana is in Room 1 1.
The yellow banana is in Room 1 2.
The red apple is in Room 1 1.
The red apple is in Room 1 2.
The diamond is oval.

This class also contains functions to activate some of the above sentences.  The sentences can be activated by
a given cell or from a spike Source.

testOneSentenceSpinn8.py, testOneSentenceNest.py, SimpleChatbotAgentSpinn8.py and
SimpleChatbotAgentNest.py use this

"""

from generatorBaseClass import GeneratorBaseClass

class GeneratorClass (GeneratorBaseClass):

    
    def __init__(self, simName, sim,neal,spinnVersion,fsa,LangFileName,OutWordsFileName):

        GeneratorBaseClass.__init__(self,simName,sim,neal,spinnVersion,fsa,LangFileName,OutWordsFileName)
        self.simName = simName
        self.sim = sim
        self.neal = neal
        self.spinnVersion = spinnVersion
        self.fsa = fsa
        self.setupGenerator()  #this builds the neural language generator topology.

 



    def TurnOnDiamondAnswerFromQuery(self,QueryCells,cell):
        """
	This connects the cell in population QueryCells to the start state of sentence "The diamond
	is oval".  Therfore, activating the cell in population QueryCells activates the sentence.        
        """
        Sentence="The diamond is oval."
        self.NeuronTurnsOnSentenceStartState(Sentence,QueryCells,cell)
        
    def TurnOnYellowBananaAnswerFromQuery(self,QueryCells,cell):
        """
	This connects the cell in population QueryCells to both the start state of sentence "The ?X is in Room ?R." and the phrase
	CAs associated to :"yellow banana" and "1 2". Therefore, activating the given cell in population QueryCells
	activates the sentence "The yellow banana is in Room 1 2."
       	"""

        Sentence="The ?X is in Room ?R."
        self.NeuronTurnsOnSentenceStartState(Sentence,QueryCells,cell)
  
        Variable="?X"
        phrase="yellow banana"  #use "NONE" if no word or phrase is required
        self.NeuronTurnsOnPhrase (phrase,Variable,QueryCells,cell)
        
        Variable="?R"
        phrase="1 2"
        self.NeuronTurnsOnPhrase (phrase,Variable,QueryCells,cell)
            

    def TurnOnRedAppleAnswerFromQuery(self,QueryCells,cell):
        """
	This connects the cell in population QueryCells to both the start state of sentence "The ?X is in Room ?R." and the
	phrase CAs associated to: "red apple" and "1 1". 
	Therefore, activating the given cell in population QueryCells activates the sentence "The red apple is in Room 1 1."
        """
        Sentence="The ?X is in Room ?R."
        self.NeuronTurnsOnSentenceStartState(Sentence,QueryCells,cell)

        Variable="?X"
        phrase="red apple"
        self.NeuronTurnsOnPhrase (phrase,Variable,QueryCells,cell)
            
        Variable="?R"
        phrase="1 1"        
        self.NeuronTurnsOnPhrase (phrase,Variable,QueryCells,cell)

    def TestSentence(self, sentence,startTime):
        """
        This activates the sentence associated with the given sentence number at the
        given startTime
        """
        if(sentence==1):
            Sentence="The ?X is in Room ?R."
            self.TurnOnSentenceFromSpikeSource(Sentence,startTime)
            Variable="?X"
            phrase="red apple"            
            self.TurnOnPhraseFromSpikeSource(phrase,Variable,startTime)
            Variable="?R"
            phrase="1 1"            
            self.TurnOnPhraseFromSpikeSource(phrase,Variable,startTime)
        elif(sentence==2):
            Sentence="The ?X is in Room ?R."
            self.TurnOnSentenceFromSpikeSource(Sentence,startTime)
            Variable="?X"
            phrase="yellow banana"            
            self.TurnOnPhraseFromSpikeSource(phrase,Variable,startTime)
            Variable="?R"
            phrase="1 2"            
            self.TurnOnPhraseFromSpikeSource(phrase,Variable,startTime)
        elif(sentence==3):
            Sentence="The diamond is oval."
            self.TurnOnSentenceFromSpikeSource(Sentence,startTime)            

    def connectQueryCellsToAnswers(self,QueryCells):
        """
        This connects the given QueryCells to the start states of the corresponding answers (sentences).
        """
        YBQuery=0  #"where is the yellow banana?" query cell
        RAQuery=1  #"where is the red apple?" query cell	
        DQuery=2   #"what shape is the diamond?" query cell
        self.TurnOnYellowBananaAnswerFromQuery(QueryCells,YBQuery)
        self.TurnOnRedAppleAnswerFromQuery(QueryCells,RAQuery)
        self.TurnOnDiamondAnswerFromQuery(QueryCells,DQuery)

