""" 
This is one of 2 files required to run the neural chatbot in nest.  The other is ChatBotInterfaceNest.py
This file (SimpleChatbotAgentNest.py) contains the chatbot (agent).  ChatBotInterfaceNest.py is the interface (world).  
You need to run these files in different shells,  run SimpleChatbotAgentNest.py in one shell
before running ChatBotInterfaceNext.py in another.
  
The agent can currently generate 3 responses:

The yellow banana is in room 1 2.
The red apple is in room 1 1.
The diamond is oval.

The language, i.e. the sentences the agent can use to answer the queries, are defined in chatbotLanguage.txt
The answer/response generated for each query (queries shown in ChatBotInterfaceNest.py) is programmed/defined in
LanguageGenerator.py.

A query entered at the interface activates a query cell in the agent (chatbot), the agent processes the query and
sends back an answer to the interface via a set of spikes.

The agent and interface communicate via three text files  outputWordsSpikes.txt and AgentQuit.txt (from agent to world) and
query.txt (from world to agent)

query.txt contains the query (i.e. in the form of an integer) from the user

outputWordsSpikes.txt contains the spikes from the agent, i.e. the spike times and the word neurons that fired.
Each word neuron represents a different word (the words represented are shown in OutputWords.txt). Therefore, the
interface builds the response (sentence) using the words in OutputWords.txt and the spikes, and spike times, from the
word neurons.

the interface will quit if AgentQuit.txt appears, i.e.when the agent (chatbot) quits



"""

import sys
import time
import os.path
import pyNN.nest as sim
import numpy as np

from nealCoverClass import NealCoverFunctions
from stateMachineClass import FSAHelperFunctions
from LanguageGenerator import GeneratorClass 

NumQueries=3  #how many queries the system will recognise

def init():
    
    sim.setup(timestep=neal.DELAY,
                    min_delay=neal.DELAY,
                    max_delay=neal.DELAY, debug=0)

def SendResponse(currentSimulationTime,WordCells,generator):
    #write to "outputWordsSpikes.txt" the word neurons that fired and the spike times
    if (currentSimulationTime == 0):
        nestOutputFileHandle = open("outputWordsSpikes.txt",'a')
        nestOutputFileHandle.write(str(currentSimulationTime)+"\n")
        nestOutputFileHandle.close()
        return 1   #has not processed query yet
    [wordSpikeTimes,wordNeuronsFired] = getWordPopSpikes(currentSimulationTime,WordCells) 
    if (len(wordNeuronsFired) > 0):
        nestOutputFileHandle = open("outputWordsSpikes.txt",'a')
        for wordNum in range (0,len(wordNeuronsFired)):
            nestOutputFileHandle.write(str(wordSpikeTimes[wordNum])+" "+                                       
                                       str(wordNeuronsFired[wordNum])+"\n")
        nestOutputFileHandle.close()


    return int(not(generator.ifFullStop(wordNeuronsFired))) # if a 1, a 0 is sent back, i.e a full response was sent


def getWordPopSpikes(currentSimulationTime,outputWordCells):
    #we extract from the spike data the word neurons that fired (and their spike times) in response to the current query
    #i.e. the neurons that fired within currentSimulationTime-30<currentSimulationTime (previous time step).

    wordNeuronsFired=[] #a record of the neurons that fired in the above interval.
    wordSpikeTimes=[]  #a record of the spike times
    wordData = outputWordCells.get_data()
    wordSegments = wordData.segments[0]
    wordSpikeTrains = wordSegments.spiketrains

    
    for wordNeuron in range (0,len(wordSpikeTrains)):
        if (len(wordSpikeTrains[wordNeuron]) > 0):
            wordSpikes = wordSpikeTrains[wordNeuron]
            for spike in range (0,len(wordSpikes)):
                SpikeTime = wordSpikes[spike]
                if (SpikeTime <= currentSimulationTime) and (SpikeTime > currentSimulationTime-30):

                    print ("Neuron-->Spike Time")  #print out the neuron that fired and the spike time.
                    print (wordNeuron,"-->",float(SpikeTime))
                    wordNeuronsFired+=[wordNeuron]
                    wordSpikeTimes+=[float(SpikeTime)]
                    
    return [wordSpikeTimes,wordNeuronsFired]

def clearQuery():
    #### this clears the query to prevent the system answering it more than once

    fileHandle = open("query.txt",'w')
    fileHandle.close()    

def ActivateQueryCell(QueryCells,Query):
    #activate the given query cell
    global NumQueries
    rest=-65.0
    firing=-45.0
    InitializeCells=np.ones(NumQueries)*rest
    InitializeCells[Query]=firing
    QueryCells.initialize(v=InitializeCells)
    
    

def RetrieveQuery(QueryCells):
    ##this retrieves the query sent by the interface, i.e. the query (integer value) contained in query.txt.
    ##It then activates the associated query cell, which then activates the response
    ProcessingQuery=0
    if (os.path.isfile("query.txt")):
        fileHandle = open("query.txt",'r')
        for inpLine in fileHandle:
            splitInp = inpLine.split()
            query = int(splitInp[0])
             #activate the queryCell associated to the given query.
            if(query==0):   
                print ("where is the yellow banana?")
                ActivateQueryCell(QueryCells,query) # activate cell 0 in QueryCell pop, which is connected to the "where is the yellow banana?"
                                                    #answer cells (sentences and phrases) 
                clearQuery()   #clear the query to prevent the system answering it more than once
                ProcessingQuery=1
            elif(query==1):
                print ("where is the red apple?")
                ActivateQueryCell(QueryCells,query)    # activate cell 1 in QueryCell pop, which is connected to the 'where is the red apple?'
                                                      #answer cells (sentences and phrases) 
                ProcessingQuery=1
                clearQuery()   
            elif(query==2):
                print ("what shape is the diamond?")
                ActivateQueryCell(QueryCells,query)    # activate cell 2 in QueryCell, which is connected to the 'what shape is the diamond?'
                                                       #answer cells (sentences and phrases) 
                ProcessingQuery=1
                clearQuery()   #clear the query to prevent the system from answering it more than once
        fileHandle.close()
        
    return ProcessingQuery

def quitInterface():
    ## simulation has finished, notify interface
    nestOutputFileHandle = open("AgentQuit.txt",'w') #create a file to let the interface know the agent has quit
    nestOutputFileHandle.close()    

def createQueryCells(generator):
    #create a pop of size NumQueries.  Each neuron corresponds to one query, activating specifc query cells will activate a specific response
    global NumQueries
    QueryCells=generator.createNeuronPop(NumQueries)  

    return QueryCells
    
def runChatBot(WordCells,QueryCells,generator):
    #this listens for a query from the interface via query.txt. Once received,
    #it processes a response to the query and sends the
    #response (i.e. the spikes from the word neurons) back via
    #outputWordsSpikes.txt.  The agent will only receive a new query when a full response has been sent 
    #to the interface, i.e. when the word neuron representing a full stop has fired.
    callbackTimeStepLength = 30
    steps = int(SIM_LENGTH/callbackTimeStepLength)
    ProcessingQuery=0   #initialisation
    for step in range (0, steps):
        currentTime = step*callbackTimeStepLength
        print ("nest simulation timestep ", currentTime)
        if(ProcessingQuery==0): #if the interface has received a full response from the agent,
                                #i.e. the neuron associated to a full stop has fired
            ProcessingQuery=RetrieveQuery(QueryCells)  #Retrieves the new query from the interface (ie in query.txt), and triggers
                                                       #the associated query cell.
        else:
            ProcessingQuery=SendResponse(currentTime,WordCells,generator) #sends the response, i.e. the spikes fired by the word population
                                                                          #within the previous time step (currentTime-30 <currentTime), to the interface
                                                                          #via outputWordsSpikes.txt
        sim.run(callbackTimeStepLength)  #runs the simulation

        time.sleep(0.7) #wait for an arbitrary time, i.e. wait for the interface to process a response

#---main 
simName = "nest"
spinnVersion = -1
LanguageFile="chatbotLanguage.txt"  #this is the file containing the chatbot's language (sentences).
OutputWordsFile="OutputWords.txt"  #where to write all of the different words comprising the language (sentences)
                                   #in the LanguageFile. 

neal = NealCoverFunctions(simName,sim,spinnVersion)
init()
fsa = FSAHelperFunctions(simName,sim,neal,spinnVersion)
generator = GeneratorClass(simName,sim,neal,spinnVersion,fsa,LanguageFile,OutputWordsFile) #setup neural language generator topology

SIM_LENGTH = 2000.0

generatorWordCells=generator.getWordCells()

#this creates the query cells that, when activated, will turn on the corresponding answer (i.e. sentence and associated phrases)
QueryCells=createQueryCells(generator)

##this connects the query cells to the corresponding answers (i.e. sentence and associated phrases)
generator.connectQueryCellsToAnswers(QueryCells)



neal.nealApplyProjections()
print (generatorWordCells)

runChatBot(generatorWordCells,QueryCells,generator)

generator.saveGenResults('results') #save the neural generator spiking results


sim.end()

##terminate interface
quitInterface() #once the simulation has finished, notify the interface

