"""
this contains all of the functions to build a neural language generation system.
you subclass this to build your own language generation topology, for an example,
see LanguageGenerator.py
"""

import numpy as np
import pickle
import os
class GeneratorBaseClass:
    
    def __init__(self, simName, sim,neal,spinnVersion,fsa,LangFileName,OutputWordsFileName):

        self.simName = simName
        self.sim = sim
        self.neal = neal
        self.spinnVersion = spinnVersion
        self.fsa = fsa
        self.LangFileName=LangFileName
        self.OutputWordsFileName=OutputWordsFileName
        self.Words=-1  #this is the set of all the different words comprising the language (sentences) in language file (LangFileName)
                       #these words are written to a text file: OutputWordsFileName
        self.WordPop=-1 #this is the population of neurons that represent the words in the self.Words array;
        		#each neuron represents one word.  
        self.StoredSenAndPhrasesInfo=-1 #this contains the sentences and the phrases from the language flie (FileName)
        self.NumStatesEachSen=-1 #this contains the number of states comprising each sentence
        self.PhraseCAs=-1        #this is a population of neurons, i.e. phrase CAs, one CA for each phrase.  When activated,
                                 #they turn on the initial state of a phrase FSA
        self.SentenceFSAs=-1	 #this is a population of neurons, i.e. sentence state CAs
        self.Phrases=0          #this states whether the language file contains phrases, default 0=no 

    def PhraseTurnsOffState(self,phrase,Variable,CAPop,CA):
        ##connects a phrase CA (associated to the given phrase and variable) to, i.e. turns off, a CA in the given population (CAPop)

        phrase=phrase+Variable ##add the variable name to the phrase

    
        PhraseCAs=self.PhraseCAs[0]
        PhraseCAIndex=self.PhraseCAs[1]

        #find the index of the phrase CA (associated to the given phrase) in the phrase CA population
        PhraseIndex=self.findPhraseCA(PhraseCAIndex,phrase)

        self.fsa.stateTurnsOffState(PhraseCAs,PhraseIndex,CAPop,CA)

    def InhibitPhraseCA (self,phrase,Variable,CAPop,CA,weight):
        ##connects a CA in the given population (CAPop) to, i.e. inhibit by a given weight, a phrase CA
        ##associated to the given phrase and variable

        phrase=phrase+Variable ##add the variable name to the phrase

    
        PhraseCAs=self.PhraseCAs[0]
        PhraseCAIndex=self.PhraseCAs[1]

        #find the index of the phrase CA (associated to the given phrase) in the phrase CA population
        PhraseIndex=self.findPhraseCA(PhraseCAIndex,phrase)

        self.fsa.stateInhibitsState(CAPop,CA, PhraseCAs,PhraseIndex, weight) ##inhibit the phrase CA so that it doesnt overexcite itself
                                                                              ##(i.e. turn itself off due to the Phrase CA's internal inhibitory
                                                                                ##neurons), especially
                                                                                ##when the phrase CA is activated by another CA
                                                                                

    def CreateWordPop(self,AllLines):
        ##create a population of neurons, where each neuron represents a different word.

        ##Number Of Sentences
        NumLines=len(AllLines)

        ##find all of the different words in the given sentences and phrases
        StoredWords=[]
        NumStoredWords=0
        for eachLine in range(0,NumLines):
            Words=AllLines[eachLine].replace('.','') ##remove . from sentence
            Words=Words.split() #split the sentence into words
        
            NumWords=len(Words)
            for eachWord in range(0,NumWords):
                alreadyrecorded=0
                if ("?" not in Words[eachWord]):
                    for eachStoredWord in range(0,NumStoredWords):
                
                        if(Words[eachWord]==StoredWords[eachStoredWord]): #if a word has been found before
                            alreadyrecorded=1
                            break

                    if(alreadyrecorded==0): #if a word has not been found before, store it and increment population size (word) counter
                        StoredWords+=[Words[eachWord]]
                        NumStoredWords+=1

        StoredWords+=["."] ##add . to the list

        ##create a CA for each word
        numInCellCA=1  #CA contains one neuron and corresponds to one word
        WordPop = self.createWordNeurons(self.fsa,NumStoredWords+1,numInCellCA) #+1 for full stop

        self.WordPop=WordPop          
        self.Words=StoredWords      


    def createWordNeurons(self,fsa,NumAssemblies,CA_SIZE):
        #create a population of word neurons
        numNeurons = CA_SIZE * NumAssemblies
        cells=self.sim.Population(numNeurons,self.sim.IF_cond_exp,self.fsa.CELL_PARAMS,label='wordsFromBoard')
        return cells

    def createNeurons(self,NumAssemblies):
        #create the given number of CAs
        numNeurons = self.fsa.CA_SIZE * NumAssemblies
        cells=self.sim.Population(numNeurons,self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        
        return cells      

    def createNeuronPop(self,neuronNum):
        #create a population of single neurons
        cells=self.sim.Population(neuronNum,self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        return cells

    def readFile(self):
        #read the contents of the language file.
        Senfile = open(self.LangFileName,"r") #opens language file
        AllLines = []
        for eachline in Senfile:
            AllLines.append(eachline.strip()) #strip gets rid of /n

        Senfile.close()
        return AllLines

    def getSentenceCells(self):
        return self.SentenceFSAs

    def ExtractSentencesAndPhrases(self,AllLines):
        ##this extracts the sentences and phrases from the given contents of a language file.
        ##return 2 lists:
        ##  First list (StoredSenAndPhrases):  contains a set of lists, the initial one contains the sentence templates and the
        ##      others contain phrase sets.  Each phrase set corresponds to a different variable, and each set contains the variable name.  we add the "NONE"
        #       phrase to each phrase list.  We use this when no word/phrase is required.  For example, if ?X does not always require a word or phrase.
        ##  Second list (NumPhrasesList) contains the number of sentences and phrases in each set.
        ##      The numbers include the variable names.

        ##Number Of Sentences
        NumLines=len(AllLines)

        StoredSenAndPhrases=[[]]
        TotalNumVars=0
        NumPhrasesList=[]
        NumPhrases=0
        totalNumSentences=1
        for eachLine in range(0,NumLines):
            Words=AllLines[eachLine].replace('.',' .') ##remove . from sentence 
            Words=Words.split()
        
            NumWords=len(Words)
            if((NumWords==1)and("?" in AllLines[eachLine])):  #IF A VARIABLE
                for VarLine in range(eachLine,NumLines):
                    if(AllLines[VarLine]!=""):
                        StoredSenAndPhrases[TotalNumVars]+=[AllLines[VarLine].replace('.',' .')]
                        NumPhrases+=1
                    else:
                        if(TotalNumVars>0):  ##do not add a NONE phrase to sentences list
                            StoredSenAndPhrases[TotalNumVars]+=["NONE"] ##add NONE to phrase list, this will be used if no word/phrase is required
                            NumPhrases+=1
                        StoredSenAndPhrases+=[[]] 
                        
                        NumPhrasesList+=[NumPhrases] #number of phrases associated to each variable (including the variable heading)
                        TotalNumVars+=1
                        NumPhrases=0
                        break

        if(StoredSenAndPhrases[1]!=[]): ##if phrases
            self.Phrases=1

        StoredSenAndPhrases.remove([]) #remove the extra []

        self.StoredSenAndPhrasesInfo=[StoredSenAndPhrases,NumPhrasesList]


    def saveGenResults(self,ResFileName):
        #save the spiking data from self.SentenceFSAs,self.WordPop,self.PhraseCAs as pkl files.
        if not os.path.exists(ResFileName):
            os.makedirs(ResFileName)
        self.SentenceFSAs.write_data(ResFileName+'/genSentence.pkl','spikes')
        self.WordPop.write_data(ResFileName+'/genWords.pkl','spikes')
        if(self.Phrases==1):
            self.PhraseCAs[0].write_data(ResFileName+'/genPhrase.pkl','spikes')

    def findStateCA(self,PhraseSenIndex,StartStateCA,nextword):
        ##return the state (CA index) associated to the position of the given word or initial state of a phrase (nextword) in a given sentence FSA, i.e.
        #return the index of the word or phrase in list PhraseSenIndex.
        #StartStateCA is the initial (index) state of a sentence FSA

        StateIndex=StartStateCA

        for i in range(0,len(PhraseSenIndex)):
            if(PhraseSenIndex[i]==" "):
                StateIndex+=1   
            if(PhraseSenIndex[i]!=nextword):
                StateIndex+=len(PhraseSenIndex[i].split()) ##number of states for a phrase equals the number of words in phrase
            else:
                break

        return StateIndex

    def findPhraseCA(self,PhraseCAList,phrase):
        ##returns the index of the phrase CA (associated to the given phrase) in the phrase CA population 
        
        StateIndex=0
        foundphrase=0
        for i in range(0,len(PhraseCAList)):
        
            if(PhraseCAList[i]==phrase):
                foundphrase=1
                break
            StateIndex+=1

        if(foundphrase==0):
            print ("ERROR: cannot find phrase in phrase list:")
            print (phrase)
            adfa
        return StateIndex

    def TurnOnPhraseFromSpikeSource (self,phrase,Variable,startTime):
        ##A spikesource turns on a phraseCA (associated to the given phrase and variable) in a specific group (Variable) 
        phrase=phrase+Variable ##add the variable name to the phrase

    
        PhraseCAs=self.PhraseCAs[0]
        PhraseCAIndex=self.PhraseCAs[1]

        #find the index of the phrase CA (associated to the given phrase) in the phrase CA population
        PhraseIndex=self.findPhraseCA(PhraseCAIndex,phrase)

        spikeGenerators = self.createOneInput(startTime)
        firstSpikeGenerator =  spikeGenerators[0]


        self.fsa.turnOnStateFromSpikeSource(firstSpikeGenerator,PhraseCAs,PhraseIndex) #turn on the phrase


    def StateTurnsOnPhrase (self,phrase,Variable,CAPop,CA):
        
        ##connects a CA in the given population (CAPop) to, i.e. activates, a phrase CA associated to the given phrase and variable
        ##phrase CA will activate the start state of a specific phrase FSA
        
        phrase=phrase+Variable ##add the variable name to the phrase

    
        PhraseCAs=self.PhraseCAs[0]
        PhraseCAIndex=self.PhraseCAs[1]

        #find the index of the phrase CA (associated to the given phrase) in the phrase CA population
        PhraseIndex=self.findPhraseCA(PhraseCAIndex,phrase)

        self.fsa.stateTurnsOnState(CAPop,CA,PhraseCAs,PhraseIndex)

            
    def NeuronTurnsOnPhrase (self,phrase,Variable,Pop,neuron):
        ##connects a neuron in the given population (Pop) to, i.e. activates, a phrase CA associated to the given phrase and variable

        ##phrase CA will activate the start state of a specific phrase FSA
        phrase=phrase+Variable ##add the variable name to the phrase

    
        PhraseCAs=self.PhraseCAs[0]
        PhraseCAIndex=self.PhraseCAs[1]

        #find the index of the phrase CA (associated to the given phrase) in the phrase CA population
        PhraseIndex=self.findPhraseCA(PhraseCAIndex,phrase)

        self.fsa.oneNeuronTurnsOnState(Pop,neuron,PhraseCAs,PhraseIndex)

    def getSpikes(self,fileName):
        #convert the data in the PklFile to spikes
        fileHandle = open(fileName,"rb")
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        NeuronsFired=[]
        SpikeTimes=[]
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    #print neuronNum, spikes[spike]
                    NeuronsFired+=[neuronNum]
                    SpikeTimes+=[float(spikes[spike])]
        fileHandle.close()

        return [NeuronsFired,SpikeTimes]

    def SortNeurons(self,WordNeuronsFired,WordSpikeTimes):
        ##sort neurons into ascending spike time order
        sort_index = np.argsort(WordSpikeTimes)
        SortedNeuronsFired=[]
        for neuron in range(0,len(WordSpikeTimes)):
                SortedNeuronsFired+=[WordNeuronsFired[sort_index[neuron]]]

        return SortedNeuronsFired

    def RetrieveWords(self):
        #retrieve the words from OutputWordsFileName
        #these are all of the different words comprising the sentences in the language file
        if (os.path.isfile(self.OutputWordsFileName)):
            OutputWords=[]
            fileHandle = open(self.OutputWordsFileName,'r')
            for inpLine in fileHandle:
                word=inpLine.split()
                OutputWords+=word
            return OutputWords
            

    def printText(self,WordNeuronsFired, Words):
        #print the sentence reported by the word neurons
        NumNeuronsFired=len(WordNeuronsFired)
        if(NumNeuronsFired>0):
            Sentence=[]
            for eachNeuron in range(0,NumNeuronsFired-1):
                Sentence+=[Words[WordNeuronsFired[eachNeuron]]]
            WholeSentence=" ".join(Sentence)+Words[WordNeuronsFired[NumNeuronsFired-1]]
        else:
            WholeSentence=""
        print (WholeSentence)

    def ifFullStop(self,WordNeuronsFired):
        #determine whether the neuron representing the full stop has fired
        fullstop=0
        NumNeuronsFired=len(WordNeuronsFired)
        if(NumNeuronsFired>0):
            for eachNeuron in range(0,NumNeuronsFired):
                if(self.Words[WordNeuronsFired[eachNeuron]]=="."):
                    fullstop=1
            
        return fullstop

    def printSentenceFromPkl(self):
        ##print the sentence reported by the word neurons after the simulation has run, i.e. from the genwords.pkl file
        [WordNeuronsFired,WordSpikeTimes]=self.getSpikes('results/genWords.pkl')
        WordNeuronsFired=self.SortNeurons(WordNeuronsFired,WordSpikeTimes)  #Sort neurons in order they spiked
        #print "Sentence From Network:"
        
        self.printText(WordNeuronsFired, self.Words)

    def printSentenceLive(self,WordNeuronsFired,WordSpikeTimes):
        ##print the sentence reported by the word neurons while the simulation is running
        print ("Response From Agent:")
        WordNeuronsFired=self.SortNeurons(WordNeuronsFired,WordSpikeTimes)  #Sort neurons in order they spiked
        self.printText(WordNeuronsFired,self.Words)

    def printSentenceLive2(self,WordNeuronsFired,WordSpikeTimes, Words):
        ##print the sentence reported by the word neurons while the simulation is running using the given words
        print ("Response From Agent:")
        WordNeuronsFired=self.SortNeurons(WordNeuronsFired,WordSpikeTimes)  #Sort neurons in order they spiked
        self.printText(WordNeuronsFired,Words)

    def getWordCells(self):
        return self.WordPop

    def createOneInput(self,startTime):
        #create a spike source
        inputSpikeTimes0 = [startTime]
        spikeArray0 = {'spike_times': [inputSpikeTimes0]}
        spikeGen0=self.sim.Population(1,self.sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')    

        return [spikeGen0]

    def TurnOnSentenceFromSpikeSource(self,Sentence,startTime):
        #this turns on the given sentence at the given time, i.e. activates the initial state of the sentence
        #FSA (associated to the given sentence) using a spike source
        SentenceOn=Sentence.replace('.',' .')
        sentenceIndex=[]
        StoredSentences=self.StoredSenAndPhrasesInfo[0][0][1:]
        NumSentences=self.StoredSenAndPhrasesInfo[1][0]-1
        foundsentence=0

        for eachSentence in range(0,NumSentences):
            if(StoredSentences[eachSentence]==SentenceOn):
                sentenceIndex=eachSentence
                foundsentence=1
                break

        if(foundsentence==0):
            print ("ERROR: could not find sentence")
            dasdf
    
        ##turn on the sentence
        spikeGenerators = self.createOneInput(startTime)
        firstSpikeGenerator =  spikeGenerators[0]

        SentenceCA=sum(self.NumStatesEachSen[:sentenceIndex])

        self.fsa.turnOnStateFromSpikeSource(firstSpikeGenerator,self.SentenceFSAs,SentenceCA)

    def StateTurnsOnSentenceStartState(self,Sentence,CAPop,CA):
        #This connects the CA in CAPop to, i.e. activates, the initial state of the sentence
        #FSA associated to the given sentence.
        
        
        SentenceOn=Sentence.replace('.',' .')
        sentenceIndex=[]
        StoredSentences=self.StoredSenAndPhrasesInfo[0][0][1:]
        NumSentences=self.StoredSenAndPhrasesInfo[1][0]-1
        foundsentence=0

        for eachSentence in range(0,NumSentences):
            if(StoredSentences[eachSentence]==SentenceOn):
                sentenceIndex=eachSentence
                foundsentence=1
                break

        if(foundsentence==0):
            print ("ERROR: could not find sentence")
            dasdf
    

        SentenceCA=sum(self.NumStatesEachSen[:sentenceIndex])

        #CA turns on the initial state of the given sentence
        self.fsa.stateTurnsOnState(CAPop,CA,self.SentenceFSAs,SentenceCA)
        

    def NeuronTurnsOnSentenceStartState(self,Sentence,Pop,Neuron):
        #This connects the Neuron in Pop to, i.e. activates, the initial state of the sentence
        #FSA associated to the given Sentence.

        SentenceOn=Sentence.replace('.',' .')
        sentenceIndex=[]
        StoredSentences=self.StoredSenAndPhrasesInfo[0][0][1:]
        NumSentences=self.StoredSenAndPhrasesInfo[1][0]-1
        foundsentence=0

        for eachSentence in range(0,NumSentences):
            if(StoredSentences[eachSentence]==SentenceOn):
                sentenceIndex=eachSentence
                foundsentence=1
                break

        if(foundsentence==0):
            print ("ERROR: could not find sentence")
            dasdf


        SentenceCA=sum(self.NumStatesEachSen[:sentenceIndex])

        #Neuron turns on the initial state of the given sentence
        self.fsa.oneNeuronTurnsOnState(Pop,Neuron,self.SentenceFSAs,SentenceCA)

    def inihibitWordNeuron(self,WordPop,WordNeuron,StatePop,FromCA):
        ##this connects FromCA in StatePop to, i.e. turns on, an inhibitory cell which inhibits the given WordNeuron in
        ##WordPop.  This is used to prevent the WordNeuron continuously firing
        InhibitCell=self.sim.Population(1,self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.fsa.stateTurnsOnOneNeuron(StatePop,FromCA,InhibitCell,0)
        connector = []
        connector=connector+[(0,WordNeuron, self.fsa.CA_STOPS_CA_WEIGHT*2.3, self.neal.DELAY)]
        self.neal.nealProjection(InhibitCell,WordPop,connector,'inhibitory')
        

    def turnOnStatesAndWordsInPhrases(self,NumPhraseInfo,StoredSenAndPhrases,nextword,StartStateCA,PhraseSenIndex,State,prevword):
        #this creates phrase FSAs for the phrases in the given phrase set (i.e. the variable in nextword), and connects each phrase FSA state
        #to a word neuron.
        #the function also connects the current part of the sentence to the next part.  That is, it connects the current state of the sentence FSA
        #(which is either a word or end of a phrase in State) to (i.e. half turns on) the initial states of the phrase FSAs.
        #A phrase CA is also connected to (i.e. half turns on) the initial state of its associated phrase FSA.

        NumPhraseVars=NumPhraseInfo[0] #number of phrase sets
        NumPhrases=NumPhraseInfo[1] #number of phrases in each set

        PhraseCAs=self.PhraseCAs[0]              ##Phrase CA population
        PhraseCAIndex=self.PhraseCAs[1]          ##this contains the phrase order, i.e.the position of each phrase in this list corresponds to
                                                 #the index of its phrase CA in the phrase CA pop

        prevstates=[]
        phrases=StoredSenAndPhrases[1:] #extract out all phrases


        for eachPhraseSet in range(0,NumPhraseVars):
            PhraseSet=phrases[eachPhraseSet] #phrase set for a particular variable

            if(nextword==PhraseSet[0]):  ##find the phrase set associated to the given variable (nextword)
                for eachPhrase in range(1,NumPhrases[eachPhraseSet]):   ##for each phrase (excluding the variable name)
                    nextPhrase=PhraseSet[eachPhrase]                                           ##get a phrase
                    NextState=self.findStateCA(PhraseSenIndex,StartStateCA,nextPhrase+PhraseSet[0]) ##find the next state CA in the SentenceFSAs CA pop,i.e.the
                                                                                            #initial state of the phrase in a given sentence.
                                                                                            ##Also tag the phrase with the associated variable name
                                                                                            ##StartStateCA is the index of the initial state of a given
                                                                                            ##sentence FSA
                                                                                            ####PhraseSenIndex contains the sentence state order for each
                                                                                            ##sentence, i.e.the position of each word and phrase in this list
                                                                                            ##corresponds to the index of its state CA in the SentenceFSAs CA
                                                                                            #pop
                                                 
           
                    if("?" in prevword):  #if the previous word was a variable (phrase)
     
                        for i in range (0,len(State)): ## connect the final state of all previous phrases to the next state, i.e. the initial state of this
                                                        #phrase
                                                       
                            self.fsa.stateHalfTurnsOnState(self.SentenceFSAs,State[i],self.SentenceFSAs,NextState)
                                                                                        ##all previous phrase FSAs will half turn on the next
                                                                                        #state, i.e. the initial state of the phrase
                                                                                        
                    else: #if previous state was a word
                        #the previous state (word) will half turn on the next state, i.e. the initial state of the phrase
                        self.fsa.stateHalfTurnsOnState(self.SentenceFSAs,State,self.SentenceFSAs,NextState)
                    
                    ####
                    PhraseState=self.findPhraseCA(PhraseCAIndex,nextPhrase+nextword) ##find the CA associated to the phrase (nextphrase) in the population of
                                                                            ##phrase CAs
                                                                            ##tag the phrase with the associated variable
                    self.fsa.stateHalfTurnsOnState(PhraseCAs,PhraseState,self.SentenceFSAs,NextState)
                                                                            ##the phrase CA that is turned on at the beginning of the sim,
                                                                            ##will half turn on the next state, i.e. the initial state of its phrase
                    self.fsa.stateTurnsOffState(self.SentenceFSAs,NextState,PhraseCAs,PhraseState)  ##turn the phraseCA off
            
                    WordsInPhrase=PhraseSet[eachPhrase].split()
                    NumWordsInPhrase=len(WordsInPhrase)
       
                    ##turn on the next state and word in a phrase                
                    for eachword in range(0,NumWordsInPhrase): 
                        self.fsa.makeCA(self.SentenceFSAs,NextState+eachword)      #make a CA for each state in a phrase
 
                        if(eachword+1<NumWordsInPhrase):
                        ##previous phrase state turns on next phrase state 
                            self.fsa.stateTurnsOnState(self.SentenceFSAs,NextState+eachword,self.SentenceFSAs,NextState+eachword+1)#the next word in
                                                                                                                    #phrase (i.e. index) is +1 after the 
                                                                                                                      #current word in phrase
                                                                                                                                                                                                                                            
                        if(WordsInPhrase[eachword]!="NONE"): ## NONE == NO WORD/PHRASE IS REQUIRED.  We do not turn on a word neuron if "NONE" is used                    
                            #turn on the associated word neuron
                            wordNeuron=self.findWordNeuron(self.Words,WordsInPhrase[eachword])
                            self.fsa.stateTurnsOnOneNeuron(self.SentenceFSAs,NextState+eachword,self.WordPop,wordNeuron)
                            if(((eachword+1)<NumWordsInPhrase)and(WordsInPhrase[eachword]!=WordsInPhrase[eachword+1])):  ##if the next word is the same as current word,
                                                                                                                        #delay inhibiting word neuron until the next word
                                                                                                     
                                                                                         
                                self.inihibitWordNeuron(self.WordPop,wordNeuron,self.SentenceFSAs,NextState+eachword) #inhibit itself otherwise the
                                                                                        #same word may be repeated (i.e. the wordneuron may continuously spike)

                        if(eachword==0):  ##if the current word is the first word in the phrase
                        ##turn off previous state (if a word) or previous final phrase states (if a phrase)
                        ###
                            if("?" in prevword): #if the previous word was the final word of a phrase, connect the initial state 
                                              #of each phrase to the previous final state of each phrase
                                             #this will turn off the previous phrase final states when one of the next phrases are triggered
                                for i in range (0,len(State)):
                                    self.fsa.stateTurnsOffState(self.SentenceFSAs,NextState+eachword,self.SentenceFSAs,State[i])  
                            else:
                                #if the previous state was a word (not a phrase) connect the initial state of each phrase to the previous word state
                                self.fsa.stateTurnsOffState(self.SentenceFSAs,NextState+eachword,self.SentenceFSAs,State)
                            ###
                        else:
                            #if current state in phrase is activated, turn off previous state in phrase (-1)
                            self.fsa.stateTurnsOffState(self.SentenceFSAs,NextState+eachword,self.SentenceFSAs,NextState+eachword-1)
    

                        if(eachword==NumWordsInPhrase-1):
                            prevstates+=[NextState+eachword]  #record the last state of each phrase

        return [prevstates,wordNeuron]

    def createPhraseCAs(self):
    #create a CA for each phrase
    #activating a phrase CA will half activate the initial state of its phrase FSA.

        StoredSenAndPhrases=self.StoredSenAndPhrasesInfo[0][1:] ##this contains lists of phrases, where each list contains a set of phrases associated
                                                       ##to a different variable
                                                       ##[0][0] contains the sentence templates
        NumPhrases=self.StoredSenAndPhrasesInfo[1][1:] #this contains the number of phrases associated to each variable.  The number also includes the
                                                       #variable headings
    
        PhraseCAIndex=[] ##this contains the phrase order, i.e.the position of each phrase in this list corresponds to
                          #the index of its phrase CA in phrase CA pop
        NumPhraseVars=len(NumPhrases)


        AllPhraseVars=len(NumPhrases)  #number of variables
        NumberPhrases=sum(NumPhrases)-AllPhraseVars  #obtain the total number of phrases.  

        PhraseCAs = self.createNeurons(NumberPhrases)
        
        for i in range(0,NumberPhrases):
            self.fsa.makeCA(PhraseCAs,i)     #make a CA 
        

        for eachPhraseSet in range(0,NumPhraseVars):
            PhraseSet=StoredSenAndPhrases[eachPhraseSet][1:] ##extract the phrases minus the variable heading
            for eachPhrase in range(0,NumPhrases[eachPhraseSet]-1):  ##-1 for the variable headings
                PhraseCAIndex+=[PhraseSet[eachPhrase]+StoredSenAndPhrases[eachPhraseSet][0]] ##attach the variable name to each phrase

        self.PhraseCAs=[PhraseCAs,PhraseCAIndex] 

    def CalculateNumStates(self):
        ##calculate the number of FSA states for each possible sentence.
        StoredSenAndPhrases=self.StoredSenAndPhrasesInfo[0] ##this contains the sentence templates and the different phrases
        NumPhrases=self.StoredSenAndPhrasesInfo[1] #this contains the number of phrases and sentence templates (including the headings)
        NumSentences=NumPhrases[0]-1 # number of sentence templates excluding the sentence heading
        NumPhrases=NumPhrases[1:]  #extract the number of phrases for each variable
        NumPhraseVars=len(NumPhrases) # number of variables (phrase sets)
        Phrases=StoredSenAndPhrases[1:] ##extract the phrases.  Each element (list) contains a phrase list associated to a different variable
        TotalNumStates=0  #total number of states for all sentences

        sentencePr=StoredSenAndPhrases[0][1:] #extract out the sentence templates but take out sentence heading

        PhraseSenIndex=[[]]
        NumStatesEachSen=np.zeros(NumSentences) ##records the number of states for each sentence template
        
        for eachsentence in range(0,NumSentences):  #each sentence template
        
            words=sentencePr[eachsentence].replace('.','') ##remove . from sentence
            words=words.split()     #split the sentence into words
            NumWords=len(words)      ##the number of words in the sentence (including variables)
            for eachWord in range(0,NumWords):
                if("?" not in words[eachWord]):  #if the word is not a variable
                    
                    TotalNumStates+=1           
                    PhraseSenIndex[eachsentence]+=[words[eachWord]]  ## add the word to the list
                    NumStatesEachSen[eachsentence]+=1   #record the number states for each sentence
                else:
                    #if word is a variable
                    for eachPhraseSet in range(0,NumPhraseVars):  #for each phrase set (i.e. for each variable)
                    
                        PhraseSet=Phrases[eachPhraseSet] #extract phrase set associated to a particular variable but exclude the variable heading

                        if(words[eachWord]==PhraseSet[0]):##find the phrase set associated to the variable name
                            if(eachWord==0):  #if the first word is a variable, we want to create a start state to trigger the sentence
                                TotalNumStates+=1
                                NumStatesEachSen[eachsentence]+=1
                                PhraseSenIndex[eachsentence]+=[" "] #this empty phrase corresponds to the start state of a sentence that begins with a variable                       
                            for eachPhrase in range(1,NumPhrases[eachPhraseSet]):  ##for each phrase
                                WordsInPhrase=PhraseSet[eachPhrase].split()
                                NumWordsInPhrase=len(WordsInPhrase)
                                TotalNumStates+=NumWordsInPhrase                      ##number of states for a phrase equals the number of words
                                NumStatesEachSen[eachsentence]+=NumWordsInPhrase
                                PhraseSenIndex[eachsentence]+=[PhraseSet[eachPhrase]+PhraseSet[0]]  ##record each phrase and tag it with the associated variable,
                                                                                                ##we tag it as we could have the same phrase associated to
                                                                                                ##different variables

            TotalNumStates+=1 #for a full stop
            NumStatesEachSen[eachsentence]+=1
            PhraseSenIndex[eachsentence]+=["."]
            PhraseSenIndex+=[[]]
    

        PhraseSenIndex.remove([])

        return [TotalNumStates,PhraseSenIndex,NumStatesEachSen]

    def CreateSentenceFSA(self,StartStateCA,NumberStates,wordsInSen,PhraseSenIndex):
        #create an FSA for one sentence.
	##For example, the sentence "The diamond is oval." has 5 states:
	##state 1: ^The diamond is oval.
	##state 2: The ^ diamond is oval.
	##state 3: The diamond ^ is oval.
	##state 4: The diamond is ^ oval.
	##state 5: The diamond is oval ^.
        #Each sentence FSA state is connected to a word neuron

        StoredSenAndPhrases=self.StoredSenAndPhrasesInfo[0] ##this contains the sentence templates and the different phrases
        NumPhrases=self.StoredSenAndPhrasesInfo[1] #this contains the number of phrases and sentence templates
                                                   #(includng the headings)
        NumSentences=NumPhrases[0] 

        NumPhrases=NumPhrases[1:]  #extract the number of phrases for each variable
        NumPhraseVars=len(NumPhrases) # number of phrase sets

        NumPhraseInfo=[NumPhraseVars,NumPhrases]
    
        State=StartStateCA          ##starting state (i.e. index of the initial state in SentenceFSAs) for the given sentence

        NumWords=len(wordsInSen)
        prevword=[]
        currentWord=wordsInSen[0]


        if("?" in currentWord): ##if first word in the sentence is a variable
            start=-1        
        else:
            start=0
    
        for eachWord in range(start,NumWords): ## for each word (and variable) in the given sentence
            #create a CA for each State in the sentence and join it to the next state, which is either a word state or the initial states
            #of phrase FSAs associated to a variable.
        
            if("?" not in currentWord):  #if the current word is not a variable           
                self.fsa.makeCA(self.SentenceFSAs,State)    ##make a CA for the current state (word) in the SentenceFSAs population
                
                wordNeuron=self.findWordNeuron(self.Words,currentWord)  # find the neuron associated to the currentword
                
                self.fsa.stateTurnsOnOneNeuron(self.SentenceFSAs,State,self.WordPop,wordNeuron) #turn on the word
                if(((eachWord+1)<NumWords)and(self.Words[eachWord]!=self.Words[eachWord+1])): ##if the next word is the same as current word,
                                                                                                # delay inhibiting word neuron until the next word
                    self.inihibitWordNeuron(self.WordPop,wordNeuron,self.SentenceFSAs,State) #inhibit itself otherwise the same word may be repeated
                                                                                                #(i.e. the wordneuron may continuously fire)

                #turn on next state 
                if((eachWord+1)<NumWords):
                    nextword=wordsInSen[eachWord+1] #next word in sentence
                    if("?" not in nextword): #if next word is not a variable
                        NextState=self.findStateCA(PhraseSenIndex,StartStateCA,nextword) ##find the state CA associated to the nextword in a given sentence
                                                                                ##StartStateCA is the index of the first state of a given sentence FSA
                        #turn on next state (word)
                        self.fsa.stateTurnsOnState(self.SentenceFSAs,State,self.SentenceFSAs,NextState)
                        prevState=State
                        State=NextState
                        
                        self.fsa.stateTurnsOffState(self.SentenceFSAs,NextState,self.SentenceFSAs,prevState)
                    else:
                        ##turn on next state, i.e. connect/turn on next phrase(s)
                        [prevStates,wordNeuron]=self.turnOnStatesAndWordsInPhrases(NumPhraseInfo,StoredSenAndPhrases,nextword,StartStateCA,PhraseSenIndex,State,currentWord)
                        State=prevStates

            else:
                #if current word is a variable (phrase)
                if(eachWord==-1):  #if the first word in the sentence is a variable
                    self.fsa.makeCA(self.SentenceFSAs,int(State))  #create start state for a sentence that starts with a variable.  This state, when
                                                                    #activated, will trigger the sentence
                                                                    
                    nextword=currentWord
                    currentWord=[]
                
                else:
                    if((eachWord+1)<NumWords):
                        nextword=wordsInSen[eachWord+1] #next word in sentence

                if("?" not in nextword): ##if the nextword is not a variable
                                     ##turn on next state and word

                    NextState=self.findStateCA(PhraseSenIndex,StartStateCA,nextword) #find the state CA associated to the next word in a given sentence
                    for i in range(0,len(State)):
                        ##use the next word state to turn off the final states of the current phrase
                        self.fsa.stateTurnsOffState(self.SentenceFSAs,NextState,self.SentenceFSAs,State[i])
                        ##use the final states of the current phrase to turn on the next state, i.e. word
                        self.fsa.stateTurnsOnState(self.SentenceFSAs,State[i],self.SentenceFSAs,NextState)
                    State=NextState

                
                
                else:
                    #if the next word is a variable, ##turn on/connect next phrase(s) associated to the nextword variable
                    [prevStates,wordNeuron]=self.turnOnStatesAndWordsInPhrases(NumPhraseInfo,StoredSenAndPhrases,nextword,StartStateCA,PhraseSenIndex,State,currentWord)
                    State=prevStates

            prevword=currentWord
            currentWord=nextword      
                      

        ##last word turns off last state (s)
        

        if("?" not in currentWord): #if currentWord is not a variable name, use the current word neuron to turn off the current sentence state
            self.fsa.oneNeuronTurnsOffState(self.WordPop, wordNeuron, self.SentenceFSAs, State)
        else:
            #if currentWord is a variable name (i.e. the current part of the sentence is a phrase), use the word neuron to turn off all the
            #final states of the current phrases
            for i in range(0,len(State)):
                self.fsa.oneNeuronTurnsOffState(self.WordPop, wordNeuron, self.SentenceFSAs, State[i])
    
    def findWordNeuron(self,WordsInPop,word):
        ##return the index of the neuron in the word population that represents the given word
        NumberCAWords=len(WordsInPop)
        wordNeuron=[]
        for eachWord in range(0,NumberCAWords):

            if(word==WordsInPop[eachWord]):
                wordNeuron=eachWord
                break

        if(wordNeuron==[]):
            print ("Word not found")
            print (word)
            error

        return wordNeuron

    def createFSASentences(self):
        ##create an FSA for each sentence


        [TotalNumStates,PhraseSenIndex,NumStatesEachSen]=self.CalculateNumStates() #calculate the number of states for each sentence template,
                                                                                   #i.e. for all possible sentences.
        self.NumStatesEachSen=NumStatesEachSen

        self.SentenceFSAs = self.createNeurons(TotalNumStates) ##contains all of the sentence FSA states

        NumSentences=self.StoredSenAndPhrasesInfo[1][0]-1  ##number of sentences (and sentence templates) in language file

        Sentences=self.StoredSenAndPhrasesInfo[0][0][1:] ##extract the sentences excluding the sentence header

        StartStateCA=0 #initial state of the first sentence

        for eachSentence in range(0,NumSentences):

            Words=Sentences[eachSentence].split()

            ##create an FSA for the sentence
            self.CreateSentenceFSA(StartStateCA,NumStatesEachSen[eachSentence],Words,PhraseSenIndex[eachSentence])
            #initial state of the next sentence FSA
            StartStateCA+=NumStatesEachSen[eachSentence]
    

    def setupRecording(self,sentenceCells,wordCells,phraseCells):
        sentenceCells.record({'spikes','v'})
        wordCells.record({'spikes','v'})
        if(self.Phrases==1):
            phraseCells.record({'spikes','v'})

    def CreateSPINNInputs(self,NumNeurons):
        ##create live spike injectors that will activate the query cells
        if(self.spinnVersion == 8):

            import spynnaker8.external_devices as externaldevices
            
            InputNeurons= self.sim.Population(
                NumNeurons, 
                externaldevices.SpikeInjector({'port':12345}),
                label='QueryToBoard')

        else:
            print ("bad spinnaker version")

        return InputNeurons

    def setupGenerator(self):
        ##this builds the neural language generator topology

        ###read in language file
    
        AllLines=self.readFile()

        ##create a population of neurons, where each neuron represents one word
        self.CreateWordPop(AllLines)

        self.WriteWordsToFile()#writes all of the words in self.Words to
                                                  #OutWordsFileName.

        ####extract the phrases and sentence templates from the langugage data
        self.ExtractSentencesAndPhrases(AllLines) 


        if(self.Phrases==1):  #if there are phrases in the given language file
            self.createPhraseCAs() ##create a CA for each phrase

        ##create an FSA for each sentence
        self.createFSASentences()
        if(self.Phrases==1):
            self.setupRecording(self.SentenceFSAs,self.WordPop,self.PhraseCAs[0])
        else:
            self.setupRecording(self.SentenceFSAs,self.WordPop,-1) #-1 for no phrases pop

    def WriteWordsToFile(self):
        #write all of the words in self.Words to file OutputWordsFileName
        WordOutputFileHandle = open(self.OutputWordsFileName,'a')
        WordNum=len(self.Words)
        for eachWord in range(0,WordNum):
            WordOutputFileHandle.write(str(self.Words[eachWord])+"\n")
        WordOutputFileHandle.close()

    def getLongestSentence(self):
        ##determine the size of the longest sentence from all possible sentences
        Sentences=self.StoredSenAndPhrasesInfo[0][0]
        NumPhraseGroups=len(self.StoredSenAndPhrasesInfo[0])-1 # -1 to exclude sentence heading
        Phrases=self.StoredSenAndPhrasesInfo[0][1:]
        NumSentences=len(Sentences)
        longestSentencesize=0
        for eachSentence in range(1,NumSentences): #start at 1 as we want to exclude the sentence heading

            Words=Sentences[eachSentence].split()
            
            NumWords=len(Words)
            numWordsInSentence=0
            for eachWord in range(0,NumWords):
                if ("?" not in Words[eachWord]):
                    numWordsInSentence+=1
                else:
                    for eachPhraseGroup in range(0,NumPhraseGroups):
                        if(Words[eachWord]==Phrases[eachPhraseGroup][0]): ##if the word in the sentence is a variable
                            NumPhrases=len(Phrases[eachPhraseGroup])
                            numWordsInPhrase=0
                            ##find the longest phrase in the phrase group
                            for eachPhrase in range (1,NumPhrases): #start at 1 to exclude phrase heading
                                WordsInPhrase=Phrases[eachPhraseGroup][eachPhrase].split()

                                if((len(WordsInPhrase)>numWordsInPhrase)and(WordsInPhrase[0]!="NONE")):
                                    numWordsInPhrase=len(WordsInPhrase)
                            numWordsInSentence+=numWordsInPhrase
                                
            if(numWordsInSentence>longestSentencesize):
                longestSentencesize=numWordsInSentence
 
        return longestSentencesize
